#include "opts.hh"
#include <stdio.h>
#include <assert.h>
#include <algorithm>
#include <iomanip>
#include <stack>

#include "verbose.hh"

#define max(a,b) (((a)>(b))?(a):(b))

//Equality and hashing for the grid
#ifdef __GNUC__
namespace __gnu_cxx 
#else
namespace std
#endif //__GNUC__
{
  template<>
  struct hash< SNode * >
  {
    size_t operator()(const SNode * s) const
    {
      return s->get_hash();
    }
  };
}; //std or __gnu_cxx
namespace std {
  template<> 
  struct equal_to< SNode * >
  {
    bool operator()(const SNode * s1, const SNode * s2) const
    {
      return *s1 == *s2;
    }
  };
}; //std


Optimist::Optimist(Architecture *_architecture)
  : architecture(_architecture), last_in(GRID_DIM)
{
  matcher = new Matcher(architecture);
  conditions = _architecture->get_conditions();
}


Optimist::~Optimist()
{
  delete matcher;
}

int Optimist::reduce_init_snode(SNode *_snode, DAG & _dag)
{
  NodeAttr *_node_attr = _snode->get_node_attr();
  vertex_set_t::iterator zi, zi_end;
  zi = _node_attr->z.begin();
  zi_end = _node_attr->z.end();
  int num_reduced = 0;

  for( ; zi != zi_end; ++zi) {
    vertex_id u_id = *zi;
    lcc::Node u = get_lcc_node(u_id, _dag);
    bool b_reduced = false;
    //For all nodes
    switch(generic(u->op)) {
    case lcc::ADDRF: case lcc::ADDRG: case lcc::ADDRL: 
//      {
//	if(u->syms[0]->sclass == lcc::REGISTER) {
//	  std::cerr << "REGISTER" << std::endl;
//	} else {
//	  std::cerr << "OTHER:" << u->syms[0]->sclass << std::endl;
//	}
//     }
      /* NO BREAK */
    case lcc::CNST:
      {
	num_reduced++;
	Matcher::match_map_iter mi, mi_end;
	//For all match of CNST, or ADDR
	for(tie(mi, mi_end) = matcher->matches(u_id, *_snode, _dag);
	    mi != mi_end; ++mi) {
	  std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;
	  
	  for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
	      iso_iter != p.second.end(); ++iso_iter) {
	    vertex_set_t chi;
	    chi.clear();
	    for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
		vi != (iso_iter->iso_map).end(); ++vi) {
	      chi.insert(vi->second);
	    }
	    //	    if(chi.size() != 1) {
	    //	      continue;
	    //	    }
	    assert(chi.size() == 1);

	    //FIXME see snode.cc
	    int height = 0;
	    const reserv_tab_t & reserv_tab = mi->first->get_reserv_tab();
	    for(int i = 0; i < reserv_tab.size(); i++) {
	      height = max(reserv_tab[i].size(), height); 
	    }
	    if(height > 0) {
	      std::cerr << "Reduce node with a reservation table!\n";
	    }
#ifdef _SP_
	    //update space profile only.
	    SpaceProfile *space_profile = _snode->get_space_profile();
	    space_profile->insert(_node_attr->alive, u_id, _dag, height,
				  mi->first->get_residence_class("dest1"), architecture);
#endif //_SP_
	    //Select only once the node
	    if(!b_reduced) {
	      _node_attr = select(chi, *_node_attr, _dag);
	      b_reduced = true;
	    }
	  }
	}
      }
    }
  }
  _snode->set_node_attr(_node_attr);
  return num_reduced;
}

Schedule *Optimist::schedule(DAG & _dag, Flags & _flags)
{
  //Statistics
  cut = improved = merged =  n_snodes = 0;
  int omega = architecture->get_omega();
  zset = new ZSet;

  //The architecture should be a parameter to schedule
  //  Schedule *s = new Schedule(omega); //extrac from SNode at the end
  NodeAttr *node_attr = get_init_node_attr(_dag);

  TimeProfile *init_time_profile = new TimeProfile(architecture->get_reserv_tab_size());
  SNode *init_snode = 
    new SNode(node_attr, init_time_profile, architecture);

  //Reduce SNode: compute ADDR and CNST nodes to their respective residences
  int num_reduced = 0; //reduce_init_snode(init_snode, _dag);

  int num_vertices = boost::num_vertices(_dag);
  coord_t coord(GRID_DIM), stop_coord(GRID_DIM);

  //Initialization
  stop_coord[T_AXIS] = 0;


  coord[T_AXIS] = 0;
  coord[L_AXIS] = 0;
  max_coord = coord;
  solution_space_t solution_space;

  solution_space.insert(init_snode, coord);
  VERB1 << "num_vertices: " << num_vertices << "\n"; //std::endl;
  VERB2 << "num_reduced: " << num_reduced << "\n"; //std::endl;
  num_vertices -= num_reduced;
  VERB2 << "num_vertices (- num_reduced): " << num_vertices << "\n"; //std::endl;
  stop_coord[L_AXIS] = num_vertices;

  for(unsigned int t = 0; t <= max_coord[T_AXIS]; t++) {

#ifndef _NOMM_
    //Memory management
    if((t > 0) && (t - 1 > 0)) { //memory criteria
      coord_t clear_v(GRID_DIM);
      clear_v[T_AXIS] = t - 1;
      std::stack< SNode * > tmp_stack; //hold pointer for deletion
      for(unsigned int l = 0; l < num_vertices; l++) {
	clear_v[L_AXIS] = l;
	grid_elem_t::iterator ci, ci_end;
	assert(tmp_stack.empty());
	for(tie(ci, ci_end) = solution_space.elements(clear_v); 
	    ci != ci_end; ++ci) {
	  tmp_stack.push(*ci);
	}
	solution_space.clear(clear_v);
	while(!tmp_stack.empty()) {
	  SNode *sn = tmp_stack.top();
	  tmp_stack.pop();
	  delete sn;
	  sn = NULL;
	}
      }
    }
#endif //_NOMM_

    //check stop
    stop_coord[T_AXIS] = t;
    if(!solution_space.empty(stop_coord)) {
      VERB1 << "\nsolution found (on entry): tau = " << t //<< std::endl
	    << "\nn_snodes: " << n_snodes
	    << " merged: " << merged
	    << " cut: " << cut
	    << " improved: " << improved << "\n"; //std::endl;
      break;
    }
    for(unsigned int l = 0; l < num_vertices; l++) {

      VERB2 << "                           \r";
      VERB2 << std::setw(3) << "l: " << std::setw(3) << l 
	    << "/" << std::setw(3) << num_vertices 
	    << std::setw(3) << " t:" << std::setw(3) << t
	    << std::setw(4) << " #n:" << std::setw(6) << n_snodes;

      coord[T_AXIS] = t;
      coord[L_AXIS] = l;
      //      std::set< SNode * >::iterator gi, gi_end;
      grid_elem_t::iterator gi, gi_end;
      for(tie(gi, gi_end) = solution_space.elements(coord);
	  gi != gi_end; ++gi) {
	SNode *current_snode =  *gi;

	//DEBUG
	/*
	VERB1 << "got elements (" << t << ", " << l << ") "
		  << (void *)current_snode << std::endl;
	VERB1 << *current_snode;
	*/
	//

	vertex_set_t::iterator zi, zi_end;
	NodeAttr *current_node_attr = current_snode->get_node_attr();
	zi = current_node_attr->z.begin();
	zi_end = current_node_attr->z.end();

	//Heuristic
	int hs = _flags.i_hs;

	for( ; zi != zi_end; ++zi) {
	  vertex_id u_id = *zi;
	  Matcher::match_map_iter mi, mi_end;
	  //For all match y covering (u_id, current_snode, & chi, & instr)
	  for(tie(mi, mi_end) = matcher->matches(u_id, *current_snode, _dag);
	      mi != mi_end; ++mi) {
	    std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;

	    //condition checking
	    int idx = 0;
	    bool cond_check;
	    cond_check = true;  //instruction without condition (default true)

	    if((idx = mi->first->get_cond_index()) != -1) {
	      cond_check = false; //instructionw with condition (default false)
	      //Apply condition check
	      pattern_t *pattern = p.first;
	      int psize = boost::num_vertices(*pattern);	 //pattern size
	      std::vector< lcc::Node > arguments(psize); //arguments(size of the pattern)
	      std::vector< iso_map_s > iso_maps = p.second;
	      int iso_vect_size = iso_maps.size();


	      //debug
// 	      std::cerr << "pattern size: " << psize << std::endl;
// 	      pvertex_iter vi_tmp, vi_tmp_end;
// 	      tie(vi_tmp, vi_tmp_end) = boost::vertices(*pattern);
// 	      std::cerr << "DUMP pattern ids nodes: ";
// 	      for(; vi_tmp != vi_tmp_end; ++vi_tmp) {
// 		std::cerr << *vi_tmp << " ";
// 	      }
// 	      std::cerr << std::endl;
	      //end dubug

	      for(int vi = 0; vi < iso_vect_size; vi++) {
		//ISO MAP
		node_map_t iso_mapp = iso_maps[vi].iso_map; 		//pattern iso_map (v -> lcc::Node)

		node_map_t::iterator bi = iso_mapp.begin(), ei = iso_mapp.end();
		for(; bi != ei; ++bi) {
		  //build the vector arguments[vi] = iso_map[vi]
		  vertex_id vid = bi->second; //DAG vertex
		  pvertex_id pvid = bi->first; //Pattern vertex
		  arguments[pvid] = get_lcc_node(vid, _dag); //map vertex_id -> lcc::Node
		  //debug
		  //		  std::cerr << "(1)arguments[" << pvid << "] = " << (void*)arguments[pvid] << std::endl;
		}

		//ARG MAP
		node_map_t arg_mapp = iso_maps[vi].arg_map; 		//pattern arg_map (v -> lcc::Node)
		bi = arg_mapp.begin();
		ei = arg_mapp.end();
		for(; bi != ei; ++bi) {
		  //build the vector arguments[vi] = iso_map[vi]
		  vertex_id vid = bi->second; //DAG vertex
		  pvertex_id pvid = bi->first; //pattern vertex
		  arguments[pvid] = get_lcc_node(vid, _dag); //map vertex_id -> lcc::Node
		  //debug
		  //		  std::cerr << "(2)arguments[" << pvid << "] = " << (void*)arguments[pvid] << std::endl;
		}

		cond_check = (conditions[idx])(arguments);
		if(cond_check) {
		  break; //Exit the consition checkni (we have a candidate)
		}
	      }
	    }
	    if(!cond_check) {
	      continue; //next candidate
	    }

	    //End of condition checking (make a method: FIXME)

	    for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
		iso_iter != p.second.end(); ++iso_iter) {
	      vertex_set_t chi;
	      chi.clear();
	      for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
		  vi != (iso_iter->iso_map).end(); ++vi) {
		chi.insert(vi->second);
	      }

	      //debug
// 	      VERB1 << *current_snode;
// 	      VERB1 << "\nFor all matches: ";
// 	      std::copy(chi.begin(), chi.end(), 
// 			std::ostream_iterator< vertex_id >(VERB1, " "));

	      update_space(current_snode, mi->first, chi, coord, solution_space, 
			   //iso_iter->root, 
			   *iso_iter, //iso_iter->arg_map, 
			   *(p.first), _dag);

	      //debug
	      /*
	      VERB1 << "\nFor all matches: ";
	      vertex_set_t & _chi = *chi_iter;
	      std::copy(_chi.begin(), _chi.end(), std::ostream_iterator< vertex_id >(VERB1, " "));
	      */
	      //

	    }
	  }

	  //Heuristic: number of variants
	  if(_flags.i_hs != 0) {
	    hs--;
	    if(hs == 0) {
	      break;
	    }
	  }

	}

#ifdef _SP_
#ifdef _TRANS_
	//Transfers
// 	SpaceProfile::iterator spi = current_snode->get_space_profile()->begin(),
// 	  spi_end = current_snode->get_space_profile()->end();

	SpaceProfile current_sp = *(current_snode->get_space_profile());
	vertex_set_t alive = current_sp.get_alive();

	int ts = _flags.i_ts;
	for(vertex_set_t::iterator a_iter = alive.begin(); 
	    a_iter != alive.end(); ++a_iter) {


	  //	for( ; spi != spi_end; ++spi) {
	  //	  vertex_id u = spi->first;
	  vertex_id u = *a_iter;
	  //	  std::set< int > & rcs = spi->second;
	  std::set< int > & rcs = current_sp[u];
	  for(std::set< int >::iterator src_iter = rcs.begin(); src_iter != rcs.end();
	      ++src_iter) {
	    std::pair< bool, std::vector< std::pair< int, TargetInstruction * > > > dests =
	      architecture->get_transfer_destinations(*src_iter);
	    if(!dests.first) {
	      continue;
	    }
	    std::vector< std::pair< int, TargetInstruction * > >::iterator dti_iter = 
	      dests.second.begin(), dti_iter_end = dests.second.end();
	    for( ; dti_iter != dti_iter_end; ++dti_iter) {
	      if(rcs.find(dti_iter->first) == rcs.end()) {
		//node u is not in residence class of the destination of the transfer instruction
		//add transfer instruction
		iso_map_s tmp_iso_maps;
		tmp_iso_maps.root = u;
		pattern_t dummy_pattern;
		update_space(current_snode, dti_iter->second, vertex_set_t(), coord, 
			     solution_space, tmp_iso_maps,
			     //u, dummy_node_map, 
			     dummy_pattern, _dag);
		if(_flags.i_ts != 0) {
		  ts--;
		  if(ts == 0) {
		    goto end_transfer;
		  }
		}
	      }
	    }
	  }
	}
      end_transfer: {}
#endif //_TRANS_
#endif //_SP_
      }
    }
    //In case the last node does not increase the total execution time
    if(!solution_space.empty(stop_coord)) {
      VERB1 << "\nsolution found (on exit): tau = " << t //<< std::endl
	    << "\nn_snodes: " << n_snodes
	    << " merged: " << merged
	    << " cut: " << cut
	    << " improved: " << improved << "\n"; //std::endl;
      break;
    }
  }

  //  std::set< SNode * >::iterator gi, gi_end;
  grid_elem_t::iterator gi, gi_end;
  tie(gi, gi_end) = solution_space.elements(stop_coord);
  if(gi == gi_end) {
    //No solution found
    VERB1 << "\nn_snodes: " << n_snodes
	  << " merged: " << merged
	  << " cut: " << cut
	  << " improved: " << improved << "\n"; //std::endl;

    assert(!solution_space.empty(last_in));

    //FIXME
//     for(tie(gi, gi_end) = solution_space.elements(last_in);
// 	gi != gi_end; ++gi) {
//       SNode *current_snode =  *gi;
//       std::cerr << *(*gi) << std::endl;
//     }

//     assert(false);
    std::cerr << "Solution not found!\n";
    exit(-1);
  }
  Schedule *s = (*gi)->get_schedule();

  //debug
//   std::cerr << "SNode solution:\n";
//   std::cerr << *(*gi);
  //

  delete init_time_profile;
  delete init_snode;
  delete zset;

  return s;
}

void Optimist::update_space(const SNode *_current_snode,
			    TargetInstruction *_target_instruction,
			    const vertex_set_t & _chi,
			    const coord_t & _current_coord,
			    solution_space_t & _solution_space,
			    //			    vertex_id & _root,
			    //			    node_map_t & _arg_map,
			    iso_map_s & _iso_maps,
			    pattern_t & _pattern,
			    DAG _dag)
{
  NodeAttr *succ_node_attr = 
    select(_chi, *_current_snode->get_node_attr(), _dag);
  SNode *succ_snode = new SNode(*_current_snode);
  succ_snode->set_node_attr(succ_node_attr);

  std::pair< bool, int > p = 
    succ_snode->insert(_chi,  //_root, 
		       _dag, _iso_maps,//_arg_map, 
		       _pattern, _target_instruction, architecture);
  //debug
//   VERB1 << "\nSched2:\n" << *(succ_snode->get_schedule());
//   VERB1 << std::endl;
  //

  if(!p.first) {
    delete succ_snode;
    cut++;
    return;
  }

  //debug
  //  VERB1 << "update_space: tinst: " << *_target_instruction << std::endl;
  //

  coord_t check_coord(GRID_DIM), dest_coord(GRID_DIM);
  check_coord = _current_coord;

  int tau = succ_snode->get_tau();
  //FIXME  assert(tau > _current_coord[T_AXIS]);
  int dest_l = _current_coord[L_AXIS] + _chi.size();
  dest_coord[L_AXIS] = dest_l;
  dest_coord[T_AXIS] = tau;

  check_coord[L_AXIS] = dest_l;
  check_coord[T_AXIS] = _current_coord[T_AXIS];
  SNode *snode_prime = NULL;
  unsigned int maxtime = max_coord[T_AXIS];
  //  unsigned int maxtime = check_coord[T_AXIS] + max_latency(funtion of moves, ISA); //FIXME

  //  std::pair< std::set< SNode *>::iterator, bool > ep;
  std::pair< grid_elem_t::iterator, bool > ep;
  for( ; check_coord[T_AXIS] <= maxtime; check_coord[T_AXIS]++) {
    ep = _solution_space.exist(succ_snode, check_coord);
    if(ep.second) {
      break;
    }
  }

  // DEBUG
  //  VERB1 << "\nsucc snode v[t:" << tau << ", l:" << dest_l << "] " << std::endl;
  //  VERB1 << (void *)succ_snode << std::endl << *succ_snode;
  //

  if(ep.second) {
    SNode *tmp_snode = *ep.first;
    assert(*tmp_snode == *succ_snode);
    if(tau < check_coord[T_AXIS]) {
      //Improved intermediate solution
      _solution_space.erase(*(ep.first), check_coord);
      SNode *sn = *(ep.first);
      delete sn;
      sn = NULL;
      _solution_space.insert(succ_snode, dest_coord);
      improved++;
    } else {
      //Newsolution is equivalent or worse (forget it)
      //Debug
//       VERB1 << "\nforgeting\n" << *succ_snode << "tau:" << succ_snode->get_tau();
//       VERB1 << "\nin favour of existing\n" << *tmp_snode 
// 		<< "tau:" << tmp_snode->get_tau() << std::endl;
      delete succ_snode;
      merged++;
    }
  } else {
    //New intermediate solution
    ep = _solution_space.insert(succ_snode, dest_coord);
    last_in = dest_coord;
    assert(ep.second); //Insertion succeeded
    n_snodes++;
    //Update max_coord
    max_coord[T_AXIS] = max(max_coord[T_AXIS], tau);
  }
}

NodeAttr *Optimist::get_init_node_attr(DAG & _dag)
{
  graph_traits< DAG >::edge_iterator ei, ei_end;
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  indeg_t indeg;
  vertex_set_t z;
	
  //All vertices start with indegree 0
  for(tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
    indeg[*vi] = 0;
  }
  //Initial indegrees
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id v_id = target(*ei, _dag);
    indeg[v_id]++;
  }	
  //Initial z
  indeg_t::iterator ii;
  for(ii = indeg.begin(); ii != indeg.end(); ++ii) {
    if(indeg[ii->first] == 0) {
      z.insert(ii->first);
    }
  }
	
  vertex_set_t alive;
	
  //Initial set of alive is empty if we only consider basic block level
  alive.clear();
  NodeAttr *node_attr = zset->insert(z, indeg, alive);
  return node_attr;
	
}

NodeAttr *Optimist::select(const vertex_set_t & _chi, const NodeAttr & _node_attr,
			   DAG & _dag)
{
  vertex_set_t z(_node_attr.z);
  indeg_t indeg(_node_attr.indeg);
  vertex_set_t alive;

  //debug
//   VERB1 << "\nselect: ";
//   std::copy(_chi.begin(), _chi.end(), std::ostream_iterator< vertex_id >(VERB1, " "));
//   VERB1 << std::endl;
  //
	
  vertex_set_t::const_iterator chi_iter = _chi.begin();
  for(; chi_iter != _chi.end(); ++chi_iter) {
    vertex_id u_id = *chi_iter;
    //successors of v_id are decremented
    successor_iterator pi, pi_end;
    for(tie(pi, pi_end) = successors(u_id, _dag);
	pi != pi_end; ++pi) {
      vertex_id v_id = *pi;
      indeg[v_id]--;
      if(indeg[v_id] == 0) {
	z.insert(v_id);
      }
    }
  }
  //Remove chi nodes from z
  for(chi_iter = _chi.begin(); chi_iter != _chi.end(); ++chi_iter) {
    assert(indeg[*chi_iter] == 0);
    z.erase(*chi_iter);
  }
  //Alive set
  graph_traits< DAG >::edge_iterator ei, ei_end;
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id u_id = source(*ei, _dag),
      v_id = target(*ei, _dag);
    if((indeg[u_id] == 0) && (z.find(u_id) == z.end())) {
      if((indeg[v_id] != 0) || (z.find(v_id) != z.end())) {
	//Target in zero-indegree set
	alive.insert(u_id);
      }
    }			
  }
	
  NodeAttr *node_attr = zset->insert(z, indeg, alive);
  return node_attr;
}

//std::pair<match_map_t::iterator, match_map_t::iterator> //Why it does not compile?
//std::pair< std::map< TargetInstruction *, std::vector< vertex_set_t > >::iterator,
//	   std::map< TargetInstruction *, std::vector< vertex_set_t > >::iterator >

// std::pair< Optimist::match_map_iter, Optimist::match_map_iter >
// Optimist::matches(vertex_id & _u_id, SNode & _current_snode, DAG & _dag)
// {
//   //build the match map
//   match_map.clear();
//   std::vector< pvertex_id > starting_nodes;
//   //LCC IR node value
//   lcc::Node u = get_lcc_node(_u_id, _dag);
//   std::pair< match_iter, match_iter > match_candidates = architecture->match_start(u->op);
//   //  < pattern_t *, TargetInstruction * > *

//   for( ; match_candidates.first != match_candidates.second; ++(match_candidates.first)) {
//     pattern_t *pattern = (*(*(match_candidates.first))).first;
//     TargetInstruction *target_instruction = (*(*(match_candidates.first))).second;
//     pvertex_iter vi, vi_end;
//     starting_nodes.clear();
//     vertex_kind_map::type vkm = vertex_kind_property(*pattern);
//     vertex_info_map::type vim = vertex_info_property(*pattern);

//     //starting nodes of the pattern (to be constructed when building/parsing) FIXME
//     pvertex_id start_node;
//     bool b_start_node = false;
//     for(tie(vi, vi_end) = vertices(*pattern); vi != vi_end; ++vi) {
//       if((vkm[*vi] == OPNODE) && vim[*vi].indeg == 0) {
// 	if(!b_start_node) {
// 	  std::list< int > & ops = vim[*vi].ops;
// 	  std::list< int >::iterator oi;
// 	  for(oi = ops.begin(); oi != ops.end(); ++oi) {
// 	    if(*oi == u->op) {
// 	      start_node = *vi;
// 	      b_start_node = true;
// 	      break;
// 	    }
// 	  }
// 	}
// 	if(start_node != *vi || !b_start_node) {
// 	  starting_nodes.push_back(*vi);
// 	}
//       }
//     } 
//     assert(b_start_node);

//     if(starting_nodes.size() == 0) {
//       std::vector< iso_map_s > iso_maps; //(1);
//       if(check(start_node, *pattern, _u_id, _dag, _current_snode, 
// 	       *target_instruction, iso_maps)) {

// 	//Add starting nodes as isomorph
// 	for(int i = 0; i < iso_maps.size(); i++) {
// 	  iso_maps[i].iso_map[start_node] = _u_id;
// 	}

// 	match_map[target_instruction] = make_pair(pattern, iso_maps);

//       }    
//     } else {
//       //All permutation on remining nodes
//       NodeAttr *node_attr = _current_snode.get_node_attr();
//       std::set< int > ops; //All op for the leave pattern
//       for(int i = 0; i < starting_nodes.size(); i++) {
// 	std::list< int > & tmp_ops = vim[starting_nodes[i]].ops;
// 	for(std::list< int >::iterator v_iter = tmp_ops.begin(); 
// 	    v_iter != tmp_ops.end(); ++v_iter) {
// 	  ops.insert(*v_iter);
// 	}
//       }
//       std::vector< vertex_id > x_i;

//       //for all nodes from z\u_id
//       for(vertex_set_t::iterator z_iter = node_attr->z.begin(); z_iter != node_attr->z.end();
// 	  ++z_iter) {
// 	if(*z_iter == _u_id) {
// 	  continue;
// 	}
// 	int op = get_lcc_node(*z_iter, _dag)->op;
// 	if(ops.find(op) != ops.end()) {
// 	  x_i.push_back(*z_iter);
// 	}
//       }
//       int start_size = starting_nodes.size();
//       if(x_i.size() < start_size) {
// 	//failed!
// 	continue;
//       }
//       //Check all permutations
//       std::sort(x_i.begin(), x_i.end());

//       std::vector< iso_map_s > base_iso_maps;
//       base_iso_maps.clear();

//       if(!check(start_node, *pattern, _u_id, _dag, 
// 		_current_snode, *target_instruction, base_iso_maps)) {
// 	continue; //for loop
//       }

//       //Add initial nodes
//       for(int i = 0; i < base_iso_maps.size(); i++) {
// 	base_iso_maps[i].iso_map[start_node] = _u_id;
//       }

//       std::vector< iso_map_s > iso_maps;
//       std::vector< std::vector< iso_map_s > >partial_iso_maps(start_size);

//       do {
// 	bool b_success = false;
// 	for(int i = 0; i < start_size; i++) {
// 	  partial_iso_maps[i].clear();
// 	  if(!check(starting_nodes[i], *pattern, x_i[i], _dag, 
// 		   _current_snode, *target_instruction, partial_iso_maps[i])) {
// 	    break; //for loop
// 	  }
// 	  //compose partial_iso_map
// 	  b_success = true;
// 	}

// 	if(b_success) {
// 	  //Concatenate base_iso_maps \bowtie parital_iso_maps
// 	  for(int i = 0; i < base_iso_maps.size(); i ++) {
// 	    //Check for same root!
// 	    vertex_id root = base_iso_maps[i].root;
// 	    for(int j = 0; j < start_size; j++) {
// 	      for(int jj = 0; jj < partial_iso_maps[j].size(); jj++) {

// 		if(partial_iso_maps[j][jj].root != root) {
// 		  //different roots (iso not equal)
// 		  continue;
// 		}
// 		//Same root
// 		iso_maps.push_back(iso_map_s());
// 		iso_map_s & current_iso = iso_maps.back();
// 		//Nodes
// 		for(node_map_t::iterator k = partial_iso_maps[j][jj].iso_map.begin(), 
// 		      k_end =  partial_iso_maps[j][jj].iso_map.end(); k != k_end; ++k) {
// 		  current_iso.iso_map[ k->first ] = k->second;
// 		}
// 		//Arguments
// 		for(node_map_t::iterator k = partial_iso_maps[j][jj].arg_map.begin(), 
// 		      k_end =  partial_iso_maps[j][jj].arg_map.end(); k != k_end; ++k) {
// 		  current_iso.arg_map[ k->first ] = k->second;
// 		}
// 		//For base nodes (start_node <-> u_id)
// 		//Nodes
// 		for(node_map_t::iterator k = base_iso_maps[i].iso_map.begin(),
// 		      k_end = base_iso_maps[i].iso_map.end(); k != k_end; ++k) {
// 		  current_iso.iso_map[ k->first ] = k->second;
// 		}
// 		//Arguments
// 		for(node_map_t::iterator k = base_iso_maps[i].arg_map.begin(),
// 		      k_end = base_iso_maps[i].arg_map.end(); k != k_end; ++k) {
// 		  current_iso.arg_map[ k->first ] = k->second;
// 		}
// 		current_iso.root = root;

// 		//Matching of current permutation
// 		for(int m = 0; m < start_size; m++) {
// 		  current_iso.iso_map[ starting_nodes[m] ] = x_i[m];
// 		}
// 	      }
// 	    }
// 	  }
// 	  //debug
// 	  VERB1 << "FOUND\n";
// 	}
//       } while(std::next_permutation(x_i.begin(), x_i.end()));

//       match_map[target_instruction] = make_pair(pattern, iso_maps);

//     }
//   }
//   //  match_map_t::iterator mi = match_map.begin(), mi_end = match_map.end();
//   match_map_iter mi = match_map.begin(), mi_end = match_map.end();

//   //debug
// //   VERB1 << "\nmatches:\n";
// //   for( ; mi != mi_end; ++mi) {
// //     for(std::vector< vertex_set_t >::iterator chi_iter = (mi->second).begin();
// // 	chi_iter != (mi->second).end(); ++chi_iter) {
// //       vertex_set_t & _chi = *chi_iter;
// //       std::copy(_chi.begin(), _chi.end(), std::ostream_iterator< vertex_id >(VERB1, " "));
// //       VERB1 << "\n---\n";
// //     }
// //   }
// //   mi = match_map.begin();
//   //

//   return std::make_pair(mi, mi_end);
// }

// Optimist::ZSet::ZSet()
// {	
// }

// Optimist::ZSet::~ZSet()
// {
//   for(z_map_t::iterator zm_iter = z_map.begin(); 
//       zm_iter != z_map.end(); ++zm_iter) {
//     assert(zm_iter->second);
//     delete zm_iter->second;
//   }
//   z_map.clear();
// }

// NodeAttr *Optimist::ZSet::get_node_attr(vertex_set_t & _z_set)
// {
//   z_map_t::iterator zm_iter = z_map.find(_z_set);
//   NodeAttr *node_attr = NULL;
//   if(zm_iter != z_map.end()) {
//     node_attr = zm_iter->second;
//   } 
//   return node_attr;
// }

// NodeAttr *Optimist::ZSet::insert(vertex_set_t & _z, indeg_t & _indeg, 
// 				 vertex_set_t & _alive)
// {
//   NodeAttr *node_attr = NULL;
//   z_map_t::iterator zm_iter = z_map.find(_z);
//   if(zm_iter == z_map.end()) {
//     //Not in the ZSet
//     node_attr = new NodeAttr(_z, _indeg, _alive);
//     z_map[_z] = node_attr;
//   } else {
//     node_attr = zm_iter->second;
//   }
//   assert(node_attr);
//   return node_attr;
// }

//Pattern matching
// bool Optimist::eq_ops_indeg(pvertex_id _u, pattern_t & _pattern, 
// 			    vertex_id _v, DAG & _dag, NodeAttr & _node_attr)
// {
//   vertex_info_map::type vim = vertex_info_property(_pattern);
  
//   if(_node_attr.indeg[_v] != vim[_u].indeg) {
//     return false;
//   }
//   std::list< int > & ops = vim[_u].ops;
//   int v_op = (get_lcc_node(_v, _dag))->op;
//   for(std::list< int >::iterator ops_iter = ops.begin(); 
//       ops_iter != ops.end(); ++ops_iter) {
//     if(*ops_iter == v_op) {
//       return true;
//     }
//   }
//   return false;
// }

// bool Optimist::eq_out_edges(pvertex_id _u, pattern_t & _pattern,
// 			    vertex_id _v, DAG & _dag, bool *_is_root)
// {
//   pedge_type_map::type petm = edge_type_property(_pattern);
//   edge_type_map::type qetm = edge_type_property(_dag);
//   int u_nb_edges[4] = { 0, 0, 0, 0 };
//   int v_nb_edges[4] = { 0, 0, 0, 0 };
//   out_pedge_iter pi, pi_end;
//   out_edge_iter qi, qi_end;

//   *_is_root = true;
//   for(tie(pi, pi_end) = out_edges(_u, _pattern); pi != pi_end; ++pi) {
//     u_nb_edges[ petm[*pi] ]++;
//     *_is_root = false;
//   }
//   for(tie(qi, qi_end) = out_edges(_v, _dag); qi != qi_end; ++qi) {
//     v_nb_edges[ qetm[*qi] ]++;
//   }
//   for(int i = 0; i < 4; i++) {
//     if(u_nb_edges[i] != v_nb_edges[i]) {
//       return false;
//     }
//   }
//   return true;
// }

// bool Optimist::eq_arg_nodes(pvertex_id & _u, pattern_t & _pattern,
// 			    vertex_id & _v, DAG & _dag,
// 			    TargetInstruction & _target_instruction,
// 			    SNode & _snode,
// 			    node_map_t & _arg_map)
// {
//   std::pair< in_pedge_iter, in_pedge_iter > p = in_edges(_u, _pattern);
//   pedge_type_map::type petm = edge_type_property(_pattern);
//   vertex_info_map::type pvim = vertex_info_property(_pattern);
//   vertex_kind_map::type pvkm = vertex_kind_property(_pattern);
//   edge_type_map::type etm = edge_type_property(_dag);
//   std::pair< in_edge_iter, in_edge_iter > q = in_edges(_v, _dag);

//   int parg_rc[2] = { -1, -1 };
//   pvertex_id kids_u[2];
//   vertex_id kids_v[2];
//   bool b_arg = false;

//   //Checking only KID0 and KID1 edges
//   for( ; p.first != p.second; ++p.first) {
//     pvertex_id kid_u = source(*(p.first), _pattern);
//     if(pvkm[kid_u] == ARGNODE) {
//       enum EDGE_T et = petm[*(p.first)];

//       //debug
// //       VERB1 << "RC : " << pvim[kid_u].id << "\n";
// //       VERB1 << "TINST:\n" << (void *)&_target_instruction 
// // 		<< " " << _target_instruction << std::endl;
//       //

//       parg_rc[et] = _target_instruction.get_residence_class(pvim[kid_u].id);
//       kids_u[et] = kid_u;
//       b_arg = true;
//     }
//   }

//   //No argument nodes
//   if(!b_arg) {
//     return true;
//   }

//   for( ; q.first != q.second; ++q.first) {
//     enum EDGE_T et;
//     switch((et = etm[*(q.first)])) {  
//     case KID0: case KID1:
//       {
// 	if(parg_rc[et] != -1) {
// 	  vertex_id kid_v = source(*(q.first), _dag);
// 	  std::pair< bool, std::set< int > > rcs = _snode.get_residence_class(kid_v);
// 	  if(!rcs.first) {
// 	    return false;
// 	  }
// 	  bool b_found = false;
// 	  //debug
// // 	  VERB1 << "RC: ";
// // 	  std::copy(rcs.second.begin(), rcs.second.end(), 
// // 		    std::ostream_iterator< int >(VERB1, " "));
// // 	  VERB1 << " - lookup: " << parg_rc[et] << std::endl;
// // 	  //
// 	  if(rcs.second.find(parg_rc[et]) != rcs.second.end()) {
// 	    b_found = true;
// 	    kids_v[et] = kid_v;
// 	    break;
// 	  }
// 	  if(!b_found) {
// 	    return false;
// 	  }
// 	}
//       }
//     }
//   }
//   //set mapping in iso_map for args
//   for(int j = 0; j < 2; j++) {
//     if(parg_rc[j] != -1) {
//       _arg_map[kids_u[j]] = kids_v[j];
//     }
//   }
//   return true;
// }

// bool Optimist::check(pvertex_id _u, pattern_t & _pattern, 
// 		     vertex_id _v, DAG & _dag, SNode & _snode,
// 		     TargetInstruction & _target_instruction,
// 		     std::vector< iso_map_s > & _iso_maps)
// {
//   node_map_t arg_map;

//   //check if opcodes are equal
//   if(!eq_ops_indeg(_u, _pattern,  _v, _dag, *(_snode.get_node_attr()))) {
//     return false;
//   }
//   //check residences of arguments
//   if(!eq_arg_nodes(_u, _pattern, _v, _dag, _target_instruction, 
// 		   _snode, arg_map)) {
//     return false;
//   }
//   //check if out-going edges are equal
//   bool is_root = false;
//   if(!eq_out_edges(_u, _pattern, _v, _dag, &is_root) || is_root)
//   {
//     if(is_root) {
//       //occurs only for one to one mapping
//       if(_iso_maps.size() == 0) {
// 	_iso_maps.push_back(iso_map_s());
//       }
//       for(int i = 0; i < _iso_maps.size(); i++) {
// 	_iso_maps[i].root = _v;
// 	_iso_maps[i].arg_map = arg_map;
//       }
//       return true;
//     }
//     return false;
//   }

//   bool b_partial_iso = false;
//   out_pedge_iter pi, pi_end;
//   pedge_type_map::type petm = edge_type_property(_pattern);  
//   std::vector< pvertex_id > p_kid0_v, p_kid1_v, p_ddep_v, p_link_v;
//   //Order pattern vertices according to their connecting edge type
//   for(tie(pi, pi_end) = out_edges(_u, _pattern); pi != pi_end; ++pi) {
//     pvertex_id p = target(*pi, _pattern);
//     switch(petm[*pi]) {

//     case KID0:
//       p_kid0_v.push_back(p);
//       break;

//     case KID1:
//       p_kid1_v.push_back(p);
//       break;

//     case DDEP:
//       p_ddep_v.push_back(p);
//       break;

//     case LINK:
//       p_link_v.push_back(p);
//       break;

//     default: 
//       assert(false);
//     }
//   }

//   std::vector< vertex_id > q_kid0_v, q_kid1_v, q_ddep_v, q_link_v;
//   out_edge_iter qi, qi_end;
//   edge_type_map::type qetm = edge_type_property(_dag);
//   for(tie(qi, qi_end) = out_edges(_v, _dag); qi != qi_end; ++qi) {
//     vertex_id q = target(*qi, _dag);
//     switch(qetm[*qi]) {
//     case KID0:
//       q_kid0_v.push_back(q);
//       break;

//     case KID1:
//       q_kid1_v.push_back(q);
//       break;

//     case DDEP:
//       q_ddep_v.push_back(q);
//       break;

//     case LINK:
//       q_link_v.push_back(q);
//       break;

//     default: 
//       assert(false);
//     }
//   }

//   std::sort(p_kid0_v.begin(), p_kid0_v.end());
//   std::sort(p_kid1_v.begin(), p_kid1_v.end());
//   std::sort(p_ddep_v.begin(), p_ddep_v.end());
//   std::sort(p_link_v.begin(), p_link_v.end());

//   std::vector< iso_map_s > partial_iso_maps;

//   //For all permutations
//   do {
//     do {
//       do {
// 	do {
// 	  for(int i = 0; i < q_kid0_v.size(); i++) {
// 	    if(!check(p_kid0_v[i], _pattern, q_kid0_v[i], _dag, _snode,
// 		      _target_instruction, partial_iso_maps)) {
// 	      goto stop_current_check;
// 	    }
// 	  }
// 	  for(int i = 0; i < q_kid1_v.size(); i++) {
// 	    if(!check(p_kid1_v[i], _pattern, q_kid1_v[i], _dag, _snode,
// 		      _target_instruction, partial_iso_maps)) {
// 	      goto stop_current_check;
// 	    }
// 	  }
// 	  for(int i = 0; i < q_ddep_v.size(); i++) {
// 	    if(!check(p_ddep_v[i], _pattern, q_ddep_v[i], _dag, _snode, 
// 		      _target_instruction, partial_iso_maps)) {
// 	      goto stop_current_check;
// 	    }
// 	  }
// 	  for(int i = 0; i < q_link_v.size(); i++) {
// 	    if(!check(p_link_v[i], _pattern, q_link_v[i], _dag, _snode,
// 		      _target_instruction, partial_iso_maps)) {
// 	      goto stop_current_check;
// 	    }
// 	  }
// 	  b_partial_iso = true;

// 	  for(std::vector< iso_map_s >::iterator pmi = partial_iso_maps.begin();
// 	      pmi != partial_iso_maps.end(); ++pmi) {
// 	    _iso_maps.push_back(iso_map_s(*pmi));
// 	    iso_map_s & part_iso = _iso_maps.back();
// 	    //extend with current matched nodes
// 	    for(int i = 0; i < q_kid0_v.size(); i++) {
// 	      part_iso.iso_map[ p_kid0_v[i] ] = q_kid0_v[i];
// 	    }
// 	    for(int i = 0; i < q_kid1_v.size(); i++) {
// 	      part_iso.iso_map[ p_kid1_v[i] ] = q_kid1_v[i];
// 	    }
// 	    for(int i = 0; i < q_ddep_v.size(); i++) {
// 	      part_iso.iso_map[ p_ddep_v[i] ] = q_ddep_v[i];
// 	    }
// 	    for(int i = 0; i < q_link_v.size(); i++) {
// 	      part_iso.iso_map[ p_link_v[i] ] = q_link_v[i];
// 	    }
// 	    //Arguments
// 	    for(node_map_t::iterator arg_iter = arg_map.begin(); 
// 		arg_iter != arg_map.end(); ++arg_iter) {
// 	      part_iso.arg_map[arg_iter->first] = arg_iter->second;
// 	    }
// 	  }

// 	stop_current_check:
// 	  //checked failed, go to next permutation
// 	  partial_iso_maps.clear();
	  
// 	} while(std::next_permutation(p_kid0_v.begin(), p_kid0_v.end()));
//       } while(std::next_permutation(p_kid1_v.begin(), p_kid1_v.end()));
//     } while(std::next_permutation(p_ddep_v.begin(), p_ddep_v.end()));
//   } while(std::next_permutation(p_link_v.begin(), p_link_v.end()));

//   return b_partial_iso;
// }


// bool Optimist::check_parents(pvertex_id _b_child, 
// 			     std::vector< pvertex_id > & _b_parents, 
// 			     pattern_t & _pattern,
// 			     TargetInstruction & _target_instruction, 
// 			     std::vector< vertex_id > & _u_parents,
// 			     vertex_set_t & _chi, DAG & _dag,
// 			     SNode & _current_snode)
// {
//   //We assume nb parents max 1, otherwise permutation. FIXME
//   //  check_this
//   if(!check_this_node(_b_parents[0], _pattern, _target_instruction, 
// 		      _u_parents[0], _dag, _current_snode)) {
//     return false;
//   }
//   //  check_other_child

//   std::pair< out_pedge_iter, out_pedge_iter > p = out_edges(_b_child, _pattern);
//   //Only one parent FIXME

//   assert(source(*p.first, _pattern) == _b_child);
//   assert(target(*p.first, _pattern) == _b_parents[0]);

//   pedge_type_map::type etm = edge_type_property(_pattern);
//   enum EDGE_T tmp_type = etm[*p.first];

//   //  check_parents;
//   return true;
// }

// bool Optimist::check_other_child(void)
// {
//   //  check_this_node;
//   //  check_children;
//   //  check_other_parents;
//   return true;
// }

// bool Optimist::check_children(void)
// {
//   //  check_this_node;
//   //  check_children;
//   return true;
// }

// bool Optimist::check_other_parents(void)
// {
//   //  check_this;
//   //  check_other_child;
//   //  check_parents;
//   return true;
// }

// bool Optimist::check_this_node(pvertex_id & _u, 
// 			       pattern_t & _pattern,
// 			       TargetInstruction & _target_instruction,
// 			       vertex_id & _v,
// 			       DAG & _dag,
// 			       SNode & _current_snode)
// {
//   vertex_kind_map::type vkm = vertex_kind_property(_pattern);
//   vertex_info_map::type vim = vertex_info_property(_pattern);
//   lcc::Node v = get_lcc_node(_v, _dag);
//   std::list< int > & ops = vim[_u].ops;
//   std::list< int >::iterator oi;
//   bool b_member = false;
//   for(oi = ops.begin(); oi != ops.end(); ++oi) {
//     if(*oi == v->op) {
//       b_member = true;
//       break;
//     }
//   }
//   //Operand in the set of operand presented in the pattern
//   if(b_member) {
//     switch(vkm[_u]) {
//     case OPNODE: {
//       //Check indeg
//       if(vim[_u].indeg != _current_snode.get_node_attr()->indeg[_v]) {
// 	return false;
//       }
//       break;
//     }
//     case ARGNODE: {
//       //Check residence class
//       char *res_class = vim[_u].id;
//       int rc_u = _target_instruction.get_residence_class(res_class);
//       int rc_v = _current_snode.get_residence_class(_v);
//       if(rc_u != rc_v) {
// 	return false;
//       }
//       break;
//     }
//     default: 
//       assert(false);
//     }
//     return true;
//   }
//   return false;
// }



// //Debug
// VERB1 << "Z_0 = { ";
// std::copy(z_indeg.z.begin(), z_indeg.z.end(), 
// std::ostream_iterator< vertex_id >(VERB1, " "));
// VERB1 << "} \nIndeg = {";
// for(ii = z_indeg.indeg.begin(); ii != z_indeg.indeg.end(); ++ii) {
// VERB1 << " (" << ii->first << "," << ii->second << ")";
// }
// VERB1 << " }\n";

// VERB1 << "CALL SELECT\n*****\n";
// z_indeg_t *new_z_indeg = zset.select(z_indeg.z, z_indeg, _dag);
// VERB1 << "AFTER CALL SELECT\n*****\n";
	
// VERB1 << "Z_0 = { ";
// std::copy(new_z_indeg->z.begin(), new_z_indeg->z.end(), 
// std::ostream_iterator< vertex_id >(VERB1, " "));
// VERB1 << "} \nIndeg = {";
// for(ii = new_z_indeg->indeg.begin(); ii != new_z_indeg->indeg.end(); ++ii) {
// VERB1 << " (" << ii->first << "," << ii->second << ")";
// }
// VERB1 << " }\n";

/* Make a boost check
//Checks (BOOST CVS)
VERB1 << "--- CHECKS ---\n";
graph_traits< DAG >::vertex_iterator vi, vi_end;
for (tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
vertex_id u_id = *vi;
VERB1 << "u_id :" << u_id << std::endl;
VERB1 << "predecessors:\n";

predecessor_iterator pi, pi_end;
for(tie(pi, pi_end) = predecessors(u_id, _dag); pi != pi_end; ++pi) {
VERB1 << *pi << " ";
}

VERB1 << "\nsuccessors:\n";
successor_iterator si, si_end;
for(tie(si, si_end) = successors(u_id, _dag); si != si_end; ++si) {
VERB1 << *si << " ";
}
VERB1 << std::endl;
}
VERB1 << "--- END CHECKS ---\n";
assert(false);
//
*/


// bool Optimist::check_match(pvertex_id & _b_u, pattern_t & _pattern, 
// 			   TargetInstruction & _target_instruction, 
// 			   vertex_id & _u, DAG & _dag, 
// 			   std::vector< vertex_set_t > & _chi_v,
// 			   SNode & _current_snode)
// {
//   //  check ARGNODES
//   std::pair< in_pedge_iter, in_pedge_iter > p = in_edges(_b_u, _pattern);
//   int kid_rc[2];
//   int nb_kids = 0;
//   pedge_type_map::type etm = edge_type_property(_pattern);
//   vertex_info_map::type vim = vertex_info_property(_pattern);
//   vertex_kind_map::type vkm = vertex_kind_property(_pattern);
//   edge_type_map::type qetm = edge_type_property(_dag);

//   //Check residence classes of ARGNODEs
//   for( ; p.first != p.second; ++(p.first)) {
//     enum EDGE_T tmp_type = etm[*(p.first)];
//     pvertex_id src_vertex = source(*(p.first), _pattern);
//     assert(vkm[src_vertex] == ARGNODE);
//     char *res_class = vim[src_vertex].id;
//     assert(tmp_type < 2);
//     kid_rc[tmp_type] = _target_instruction.get_residence_class(res_class);
//     nb_kids++;
//   }

//   std::pair< in_edge_iter, in_edge_iter > q = in_edges(_u, _dag);
//   for( ; q.first != q.second; ++(q.first)) {
//     enum EDGE_T tmp_type = get_edge_type(*(q.first), _dag);
//     vertex_id v = source(*(q.first), _dag);
//     assert(tmp_type < 2);
//     std::set< int > & rcs = _current_snode.get_residence_class(v);
//     if(rcs.find(kid_rc[tmp_type]) == rcs.end()) {
//       return false;
//     }
//   }

//   std::vector< iso_map_s > iso_maps(1);
//   if(!check(_b_u, _pattern, _u, _dag, *_current_snode,
// 	    _target_instruction, iso_maps)) {
//     return false;
//   } else {
//     //Extend with matching _b_u -> _u
//     if(iso_maps.size() == 0) {
//       //single node
//       iso_maps.push_back(iso_map_s());
//     }
//     for(std::vector< iso_map_s >::iterator imi = iso_maps.begin();
// 	imi != iso_maps.end(); ++imi) {
//       (*imi).iso_map[_b_u] = _u;
//     }
//   }

//   assert(iso_maps.size() > 0);
//   //iso maps contain all isomorphisms pattern_vertex -> dag_vertex
//   for(int i = 0; i < iso_maps.size(); i++) {
//     vertex_set_t tmp_chi;

//     //debug
//     VERB1 << "\nIsomorphism: ";
//     //

//     for(node_map_t::iterator iter = (iso_maps[i]).iso_map.begin(); 
// 	iter != (iso_maps[i]).iso_map.end(); ++iter) {
//       tmp_chi.insert(iter->second);

//       //debug
//       VERB1 << iter->first << " -> " << iter->second << " ";
//       //

//     }

//     //FIXME: check should check the RC for ARGNODEs of nodes in the pattern
//     //Here we need only to map the name of argnode to <vertex_id, RC> tuple
//     //std::map< char *, std::pair< vertex_id, RC:int > >, thus when looking for
//     //the instance of the operand, we request it from the
//     //SpaceProfile->get_rc_instance(pair<vertex_id, RC>); We need to send the vertex_id
//     //and residence class index, since a same node can be alive in different RCs
//     //simultanouesly.
//     bool b_chi_exists = false;
//     for(std::vector< vertex_set_t >::iterator chi_iter = _chi_v.begin();
// 	chi_iter != _chi_v.end(); ++chi_iter) {
//       if((*chi_iter) == tmp_chi) {
// 	b_chi_exists = true;
// 	break;
//       }
//     }

//     if(b_chi_exists) {
//       tmp_chi.clear();
//       //debug
//       VERB1 << "\ntmp_chi: exists already\n";
//       //
//       continue;
//     }

//     //Build implicitly
//     //Create new map FIXME - check

//     std::map< char *, std::pair< vertex_id, int >, ltstr > argnode_map;
//     for(std::map< pvertex_id, vertex_id >::iterator iter = (iso_maps[i]).begin(); 
// 	iter != (iso_maps[i]).end(); ++iter) {
     
//       for(p = in_edges(iter->first, _pattern); p.first != p.second; ++(p.first)) {
// 	pvertex_id src_vertex = source(*(p.first), _pattern);
// 	if(vkm[src_vertex] == ARGNODE) {
// 	  enum EDGE_T kid_type = etm[*(p.first)];
// 	  vertex_id q = iter->second;
// 	  for(std::pair< in_edge_iter, in_edge_iter > pq = in_edges(q, _dag);
// 	      pq.first != pq.second; ++(pq.first)) {
// 	    if(kid_type == qetm[*(pq.first)]) {
// 	      //id of the arg node in the DAG
// 	      vertex_id dag_arg_node = source(*(pq.first), _dag);
// 	      char *arg_node_name = vim[src_vertex].id;
// 	      int rc = _target_instruction.get_residence_class(arg_node_name);

// 	      //insert into map FIXME check for existance
// 	      argnode_map[arg_node_name] = std::make_pair(dag_arg_node, rc);
	      
// 	      break;
// 	    }
// 	  }
// 	}
//       }
//     }


//     //Add new chi
//     //_chi_v.push_back(vertex_set_t(tmp_chi));


//     vertex_set_t & current_chi = _chi_v.back();
//     //debug 
//     VERB1 << "\nIsomorphism: ";
//     //
//     for(std::map< pvertex_id, vertex_id >::iterator iter = (iso_maps[i]).begin(); 
// 	iter != (iso_maps[i]).end(); ++iter) {
//       current_chi.insert(iter->second);
//       //debug
//       VERB1 << iter->first << " -> " << iter->second << " ";
//       //
//     }

//   }

//   return true;
// }

    /*
    for(int i = 0; i < starting_nodes.size(); i++) {
      pvertex_id b_u = starting_nodes[i];
      std::vector< vertex_set_t > chi_v;
      chi_v.clear();
      if(check_match(b_u, *pattern, *target_instruction, _u_id, chi_v, 
      		     _dag, _current_snode)) {
      	match_map[target_instruction] = chi_v;
      }

      //      chi.insert(_u_id); //FIXME - for testing
      //      match_map[target_instruction] = chi; //FIXME - for testing
   
    }
   */
// 	//debug
// 	for(int i = 0; i < iso_maps.size(); i++) {
// 	  VERB1 << "iso: ";
// 	  for(node_map_t::iterator nmi = iso_maps[i].iso_map.begin();
// 	      nmi != iso_maps[i].iso_map.end(); ++nmi) {
// 	    VERB1 << nmi->first << "->" << nmi->second << " ";
// 	  }
// 	  VERB1 << "\narg: ";
// 	  for(node_map_t::iterator nmi = iso_maps[i].arg_map.begin();
// 	      nmi != iso_maps[i].arg_map.end(); ++nmi) {
// 	    VERB1 << nmi->first << "->" << nmi->second << " ";
// 	  }
// 	  VERB1 << "\nroot: " << iso_maps[i].root << std::endl;
// 	}
	//

// if(check_match(start_node, *pattern, *target_instruction, _u_id,
// 		     _dag,  chi_v, _current_snode)) {
// 	match_map[target_instruction] = chi_v;
//
// 	//debug
// 	VERB1 << "\nchi_v.size(): " << chi_v.size() << std::endl;
// 	//
//    }
